#include <stdio.h>
#include <string.h>
#include <cinttypes>
#include "mypredictor.h"
#include <climits>
#include <stdlib.h>
#include <string.h>
#include <cassert>
#include <cstdio>
#include <math.h>

#define MAX_CONF 25            // Saturating Confidence value
#define MAX_PREF 7             // Saturating Preference Counter value
#define FOLD_AMT 18            // Number of bits to fold for hashing
#define THRESHOLD -20          // Threshold value for blacklisting
#define MAX_SCORE 30           // Saturating score value for blacklisting
uint64_t MAX_SPACE = 12884901888;// Space limit (12GB) to reduce memory footprint

uint64_t curr_seq_no;       // Represent the latest seq_no for which update predictor has been called
int curr_seq_index;

static DFCM dfcm;



// Global branch and path history
static uint64_t ghr = 0, phist = 0;

// Load/Store address history
static uint64_t addrHist = 0;
// True when DFCM does the prediction
bool predicted_by_me = false;
// Maintains the total space used so far. Initialized with the size of static buffers like arrays.
uint64_t space = 595000;

uint64_t min(uint64_t a, uint64_t b)
{
    if (a < b)
        return a;
    else
        return b;
}

// XORs 'bits' most significant bits of 'result' to itself and chop them off from the front.
uint64_t fold(uint64_t result, int bits)
{
   if (!bits)
       return result;

   uint64_t val = result;
   val ^= (result >> (64 - bits));

   return val & (((uint64_t)1 << (64 - bits)) - 1);
}

// For prediction by DFCM
// Returns true if prediction is done and the predicted value is stored in predicted_value.
bool DFCM::getPrediction(uint64_t seq_no, uint64_t pc, uint8_t piece, uint64_t& predicted_value)
{
    // Accessing the first table. A different entry will be accessed for each piece.
    // Instructions are 4-bytes, so shift PC by 2.
    uint64_t firstLevelIndex = ((pc << 1) ^ piece);
    predicted_by_me = false;

    // If PC is never seen before then dont predict and add it to the table.
    std::unordered_map<uint64_t, FirstLevelEntry>::iterator i = firstLevelTable.find(firstLevelIndex);
    if(i == firstLevelTable.end())
    {
        space += (sizeof(FirstLevelEntry) + 8);
        if (space >= MAX_SPACE)
        {
            // remove a random entry from the table before inserting a new entry
            auto random_it = std::begin(firstLevelTable);
            firstLevelTable.erase(random_it);
            space -= (sizeof(FirstLevelEntry) + 8);
        }
        firstLevelTable[firstLevelIndex] = FirstLevelEntry();
        return false;
    }

    // getting indexes for all the second level tables.
    std::unordered_map<uint64_t, SecondLevelEntry>::iterator ip, other;
    uint64_t secondLevelIndex[TABLES];
    for (int j = 0; j < TABLES; j++)
    {
        secondLevelIndex[j] = i->second.predictTimeIndex[j];
    }

    // getting the table with maximum preference for prediction.
    int max_pref = -MAX_PREF;
    int max_table = 0;
    for (int j = 0; j < TABLES; j++)
    {
        if (i->second.preference[j] > max_pref)
        {
            max_pref = i->second.preference[j];
            max_table = j;
        }
    }

    // if entry doesn't exist, then add it and return false for prediction.
    ip = secondLevelTable[max_table].find(secondLevelIndex[max_table]);
    if(ip == secondLevelTable[max_table].end())
    {
        // insert the entry in all the tables if doesn't exist.
        for (int i = 0; i < TABLES; i++)
        {
            ip = secondLevelTable[i].find(secondLevelIndex[i]);
            if (ip == secondLevelTable[i].end())
            {
                space += (sizeof(SecondLevelEntry) + 8);
                if (space >= MAX_SPACE)
                {
                    // remove a random entry from the table before inserting a new entry
                    auto random_it = std::begin(secondLevelTable[i]);
                    secondLevelTable[i].erase(random_it);
                    space -= (sizeof(SecondLevelEntry) + 8);
                }
                secondLevelTable[i][secondLevelIndex[i]] = SecondLevelEntry();
            }
        }
        return false;
    }

    // insert the entry in all the other tables if doesn't exist.
    for (int i = 0; i < TABLES; i++)
    {
        other = secondLevelTable[i].find(secondLevelIndex[i]);
        if (other == secondLevelTable[i].end())
        {
            space += (sizeof(SecondLevelEntry) + 8);
            if (space >= MAX_SPACE)
            {
                //return false;
                    auto random_it = std::begin(secondLevelTable[i]);
                //auto random_it = std::next(std::begin(secondLevelTable[i]), rand()%(secondLevelTable[i].size()));
                secondLevelTable[i].erase(random_it);
                space -= (sizeof(SecondLevelEntry) + 8);
            }
            secondLevelTable[i][secondLevelIndex[i]] = SecondLevelEntry();
        }
    }

    // Don't predict if predictValue is false or maximum preference is negative.
    if (firstLevelTable[firstLevelIndex].predictValid == false || max_pref < 0)
    {
        lastStride = 0;
        return false;
    }

    // if the confidence is saturated, then we can make the prediction
    uint8_t confidence = ip->second.conf;
    if (confidence >= i->second.maxConf)
    {
        predicted_value = i->second.predictValue + ip->second.pred;
        lastPrediction = i->second.predictValue + ip->second.pred;
        lastStride = ip->second.pred;
        predicted_by_me = true;
        return true;
    }

    // else just get the predicted value for increasing the confidence at commit time, but return false
    predicted_value = i->second.baseValue + ip->second.pred;
    lastPrediction = i->second.baseValue + ip->second.pred;
    lastStride = ip->second.pred;

    // Speculate using the prediction only if confidence is high enough
    return confidence >= i->second.maxConf;
}

void DFCM::speculativeUpdate(uint64_t seq_no, bool eligible, uint8_t prediction_result, uint64_t pc,
		       uint64_t next_pc, InstClass insn, uint8_t piece, uint64_t src1, uint64_t src2,
		       uint64_t src3, uint64_t dst)
{
    bool isPredictable = 1;

    uint64_t firstLevelIndex = ((pc << 1) ^ piece);

    // finding indexes in all the tables.
    std::unordered_map<uint64_t, FirstLevelEntry>::iterator i = firstLevelTable.find(firstLevelIndex);
    uint64_t secondLevelIndex[TABLES];
    for (int j = 0; j < TABLES; j++)
    {
        secondLevelIndex[j] = i->second.predictTimeIndex[j];
    }

    // It's an instruction we are interested in predicting, update the first table history
    // Note that some other type of predictors may not want tupdate at this time if the
    // prediction is unknown to be correct or incorrect
    if (isPredictable)
    {
        // push into the list of inflight instructions
        inflightPreds.push_front({seq_no, firstLevelIndex, secondLevelIndex});
        firstLevelTable[firstLevelIndex].inflight++;
        uint64_t result = lastStride;

        // find the table which is used for prediction
        int max_pref = -MAX_PREF;
        int max_table = 0;
        for (int j = 0; j < TABLES; j++)
        {
            if (i->second.preference[j] > max_pref)
            {
                max_pref = i->second.preference[j];
                max_table = j;
            }
        }

        firstLevelTable[firstLevelIndex].predictValue = lastPrediction;

        // hashing the values to get the index of second level table.
        for (int j = 0; j < TABLES; j++)
        {
            firstLevelTable[firstLevelIndex].predictTimeIndex[j] <<= (uint64_t)pow(2, TABLES - j - 1);
            //folding into half.
            result = fold(result, FOLD_AMT);
            firstLevelTable[firstLevelIndex].predictTimeIndex[j] ^= result;
        }
    }

    // At this point, any branch-related information is architectural, i.e.,
    // updating the GHR/LHRs is safe.
    bool isCondBr = insn == condBranchInstClass;
    bool isIndBr = insn == uncondIndirectBranchInstClass;

    // Infrastructure provides perfect branch prediction.
    // As a result, the branch-related histories can be updated now.
    if(isCondBr)
        ghr = (ghr << 1) | (pc != next_pc + 4);

    if(isIndBr)
        phist = (phist << 4) | (next_pc & 0x3);
}

void DFCM::updatePredictor(uint64_t seq_no, uint64_t actual_addr,
		     uint64_t actual_value, uint64_t actual_latency)
{
    // getting the element from the list of inflights.
    std::deque<DFCM::InflightInfo> & inflight = inflightPreds;

    // If we have value predictions waiting for corresponding update
    if(inflight.size() && seq_no == inflight.back().seqNum)
    {
        // It is now safe to update the DFCM value history
        uint64_t commitTimeSecondLevelIndex[TABLES];
        for (int i = 0; i < TABLES; i++)
            commitTimeSecondLevelIndex[i] =
                firstLevelTable[inflight.back().firstLevelIndex].commitTimeIndex[i];

        // stride
        uint64_t result =
            actual_value - firstLevelTable[inflight.back().firstLevelIndex].baseValue;

        // find the hash for commit time indexing
        for (int i = 0; i < TABLES; i++)
        {
            firstLevelTable[inflight.back().firstLevelIndex].commitTimeIndex[i] <<= (uint64_t)pow(2, TABLES - i - 1);
            //folding into half.
            result = fold(result, FOLD_AMT);
            firstLevelTable[inflight.back().firstLevelIndex].commitTimeIndex[i] ^= result;
        }

        // If there are not other predictions corresponding to this PC in flight in the pipeline,
        // we make sure the DFCM value history used to predict is clean by copying the architectural value
        // history in it.
        // Speculative DFCM value history can become state when it is updated in _speculativeUpdate() with a prediction
        // that we don't know the outcome of yet, because it was not used to speculate.
        if(--firstLevelTable[inflight.back().firstLevelIndex].inflight == 0)
        {
            for (int i = 0; i < TABLES; i++)
            {
                firstLevelTable[inflight.back().firstLevelIndex].predictTimeIndex[i] =
                    firstLevelTable[inflight.back().firstLevelIndex].commitTimeIndex[i];
                firstLevelTable[inflight.back().firstLevelIndex].predictValue = actual_value;
                firstLevelTable[inflight.back().firstLevelIndex].predictValid = true;
            }
        }

        // update the confidence and the preference for all the tablels
        // also update the strides
        for (int i = 0; i < TABLES; i++)
        {
            if (secondLevelTable[i].find(commitTimeSecondLevelIndex[i]) == secondLevelTable[i].end()) {
                space += (sizeof(SecondLevelEntry) + 8);
                if (space >= MAX_SPACE)
                {
                    // remove a random entry from the table before inserting a new entry
                    auto random_it = std::next(std::begin(secondLevelTable[i]), rand()%(secondLevelTable[i].size()));
                    secondLevelTable[i].erase(random_it);
                    space -= (sizeof(SecondLevelEntry) + 8);
                }
                secondLevelTable[i][commitTimeSecondLevelIndex[i]] = SecondLevelEntry();
            }

            if (secondLevelTable[i][commitTimeSecondLevelIndex[i]].pred ==
                    actual_value - firstLevelTable[inflight.back().firstLevelIndex].baseValue)
            {
                (firstLevelTable[inflight.back().firstLevelIndex].preference[i] < MAX_PREF) ?
                    (firstLevelTable[inflight.back().firstLevelIndex].preference[i]++) :
                    firstLevelTable[inflight.back().firstLevelIndex].preference[i];
                secondLevelTable[i][commitTimeSecondLevelIndex[i]].conf =
                min(secondLevelTable[i][commitTimeSecondLevelIndex[i]].conf + 1, MAX_CONF);
            }
            // only remove the entry if it's Confidence is already 0
            else if (secondLevelTable[i][commitTimeSecondLevelIndex[i]].conf == 0)
            {
                secondLevelTable[i][commitTimeSecondLevelIndex[i]].pred =
                    actual_value - firstLevelTable[inflight.back().firstLevelIndex].baseValue;
                (firstLevelTable[inflight.back().firstLevelIndex].preference[i] > -MAX_PREF) ?
                    (firstLevelTable[inflight.back().firstLevelIndex].preference[i]--) :
                    firstLevelTable[inflight.back().firstLevelIndex].preference[i];
            }
            else
            {
                secondLevelTable[i][commitTimeSecondLevelIndex[i]].conf = 0;
                (firstLevelTable[inflight.back().firstLevelIndex].preference[i] > -MAX_PREF) ?
                    (firstLevelTable[inflight.back().firstLevelIndex].preference[i]--) :
                    firstLevelTable[inflight.back().firstLevelIndex].preference[i];
            }
        }

        firstLevelTable[inflight.back().firstLevelIndex].baseValue = actual_value;
        inflight.pop_back();
    }

    // It is now safe to update the address history register
    //if(insn == loadInstClass || insn == storeInstClass)
    if(actual_addr != 0xdeadbeef)
        addrHist = (addrHist << 4) | actual_addr;
}

// Add information of a PC to the PCIT if not present already
void addPCITEntry(uint64_t pc, uint8_t piece, uint64_t src1, uint64_t src2, uint64_t src3, uint64_t dst, InstClass insn, bool eligible)
{
    PCITEntry pcitEntry;

    std::unordered_map<uint64_t, PCITEntry>::iterator i = PCIT.find((pc << 1) ^ piece);

    // only add if the PC is not already present in the table
    if (i == PCIT.end())
    {
        pcitEntry.src[0] = src1;
        pcitEntry.src[1] = src2;
        pcitEntry.src[2] = src3;
        pcitEntry.dst = dst;
        pcitEntry.imm = 0;
        pcitEntry.op = Init;
        pcitEntry.type = none;
        pcitEntry.sure = 0;
        pcitEntry.insn = insn;
        pcitEntry.eligible = eligible;
        pcitEntry.lastdata1.valid = false;
        pcitEntry.lastdata2.valid = false;
        pcitEntry.lastdata3.valid = false;
        pcitEntry.score = 0;
        PCIT[(pc << 1) ^ piece] = pcitEntry;

        space += (sizeof(PCITEntry) + 8);
        if (space >= MAX_SPACE)
        {
            // remove a random entry from the table before inserting a new entry
            auto random_it = std::begin(PCIT);
            space -= (sizeof(PCITEntry) + 8);
            PCIT.erase(random_it);
        }
    }
    else
    {
        pcitEntry = i->second;
    }
}

// Adds a new entry to the Register File Buffer and invalidates the old entry
// also updates the SeqNotoPCMapper and RFBIndextoSeqNoMapper accordingly
int addRF(uint64_t seq_no)
{
    // curr_sequence_index represents the index of the entry to be
    // evicted, and is updated in the FIFO order
    SeqNotoPCMapper[seq_no].index = curr_seq_index;
    if (RFBIndextoSeqNoMapper[curr_seq_index] != ULONG_MAX) {
        //SeqNotoPCMapper[RFBIndextoSeqNoMapper[curr_seq_index]].index = -1;
        SeqNotoPCMapper.erase(RFBIndextoSeqNoMapper[curr_seq_index]);
    }
    RFBIndextoSeqNoMapper[curr_seq_index] = seq_no;
    curr_seq_index = (curr_seq_index + 1) % MAX_SEQ_BUFFER;
    return SeqNotoPCMapper[seq_no].index;
}

// Update the register file buffer with new register values or invalidate the entry
void updateRF(uint64_t index, uint64_t reg, uint64_t value, bool valid, uint64_t seq_no, uint64_t piece)
{
    int temp_index = SeqNotoPCMapper[index].index;
    assert(temp_index >= 0 && temp_index < MAX_SEQ_BUFFER);

    // The seq_no of the instruction that is writing to the register file entry
    // should be greater than the instruction that has already written to the
    // register entry.
    if (RegisterFileBuffer[temp_index][reg].seq_no > seq_no)
        return;

    RegisterFileBuffer[temp_index][reg].value = value;
    RegisterFileBuffer[temp_index][reg].valid = valid;
    RegisterFileBuffer[temp_index][reg].seq_no = seq_no;
    RegisterFileBuffer[temp_index][reg].piece = piece;
}

// Copy a register file entry from src sequence number to the dst sequence number
void copyRF(uint64_t src, uint64_t dst)
{
    int src_index = SeqNotoPCMapper[src].index;
    int dst_index;

    if (src == 0)
         src_index = addRF(src);

    assert(src_index >= 0);
    if (SeqNotoPCMapper.find(dst) == SeqNotoPCMapper.end())
        dst_index = addRF(dst);
    else
        dst_index = SeqNotoPCMapper[dst].index;

    memcpy(RegisterFileBuffer[dst_index], RegisterFileBuffer[src_index], MAX_REGS*sizeof(struct RegisterFileEntry));
}

// Estimate the value of the instruction with the given seq_no and store the value in
// predicted_value parameter. Returns true if we were able to estimate it
bool speculate(uint64_t seq_no, uint8_t piece, uint64_t& predicted_value)
{
    PCITEntry pcitEntry;
    uint8_t num_src = 0;
    uint64_t src_val[3];
    if (SeqNotoPCMapper.find(seq_no) == SeqNotoPCMapper.end())
        return false;

    int index = SeqNotoPCMapper[seq_no].index;

    if (index == -1)
        return false;
    assert(index >= 0 && index < MAX_SEQ_BUFFER);

    std::unordered_map<uint64_t, PCITEntry>::iterator i = PCIT.find(SeqNotoPCMapper[seq_no].pc);

    // no prediction if PC is not in the table
    // it will be added to the table in speculative update
    if (i == PCIT.end())
        return false;

    // if opcode is none or init, then no prediction
    // if surity of opcode is less than 2 then no prediction
    // In both the cases we have not learned the operation
    if (i->second.op == None || i->second.op == Init || i->second.sure < 2)
        return false;

    pcitEntry = i->second;

    // find the values of source register in register file to calculate the predicted value
    for (int i = 0; i < 3; i++)
    {
        // Indicates that the instruction is using this register, 0xdeadbeef shows that this
        // register was not used in this instruction
        if (pcitEntry.src[i] != 0xdeadbeef)
        {
            // if register file entry is valid for that source register then take it directly from there
            // otherwise recursively compute the value of that source register using speculate
            if (!RegisterFileBuffer[index][pcitEntry.src[i]].valid)
            {
                // If the sequence no of the instruction is less than the last commited instruction than
                // its value should have been commited and will be reflected in the register file entry of
                // the last commited instruction
                if (RegisterFileBuffer[index][pcitEntry.src[i]].seq_no < (curr_seq_no))
                {
                    // The register can be invalid even after the last commited instruction because of
                    // in-eligible instructions, as for them we know the destination register that they
                    // are writing to, but not the value
                    if (RegisterFileBuffer[SeqNotoPCMapper[curr_seq_no + 1].index][pcitEntry.src[i]].valid)
                    {
                        src_val[i] = RegisterFileBuffer[SeqNotoPCMapper[curr_seq_no+1].index][pcitEntry.src[i]].value;
                        continue;
                    }

                    return false;
                }
                // compute the value of the register recursively
                // if not found then we can't predict using estimator
                if (!speculate(RegisterFileBuffer[index][pcitEntry.src[i]].seq_no,
                               RegisterFileBuffer[index][pcitEntry.src[i]].piece,
                               src_val[i]))
                {
                    return false;
                }
            }
            // pick the value directly from the register file
            // this is the case when the value in the register file is valid
            else
            {
                src_val[i] = RegisterFileBuffer[index][pcitEntry.src[i]].value;
            }
            num_src++;
        }
    }

    // We got the source register value and the opcode
    // time to get the predicted_value
    // call calculate with source register values and opcode
    Data data;
    data.src_val[0] = src_val[0];
    data.src_val[1] = src_val[1];
    data.src_val[2] = src_val[2];

    predicted_value = calculate(pcitEntry.op, pcitEntry.type, data, pcitEntry.imm);

    return true;
}

bool getPrediction(uint64_t seq_no, uint64_t pc, uint8_t piece, uint64_t& predicted_value)
{
    // InsMapping is use to carry information of inflight instructions
    InsMapping insMapping;
    // Initialising variables
    if (SeqNotoPCMapper.find(seq_no) == SeqNotoPCMapper.end())
        SeqNotoPCMapper[seq_no].index = -1;
    SeqNotoPCMapper[seq_no].pc = (pc << 1) ^ piece;
    insMapping.pc = pc;
    insMapping.piece = piece;
    insMapping.seq_no = seq_no;
    insMapping.predicted = false;
    uint64_t ld_predict = 0;        // Predicted value by DFCM
    // Estimator gets the prediction by dfcm
    bool loaded = dfcm.getPrediction(seq_no, pc, piece, ld_predict);

    // Then tries to predict the instruction itself
    // the below 2 ifs cover the case when we are seeing the instruction for the first time
    // So we won't predict using dfcm as it's not trained on this PC yet
    std::unordered_map<uint64_t, PCITEntry>::iterator i = PCIT.find(SeqNotoPCMapper[seq_no].pc);

    if (i == PCIT.end())
    {
        InFlightInsL1.push(insMapping);
        return false;
    }

    // speculate tries to speculate the value of instruction based on its previous behaviour
    bool speculated = speculate(seq_no, piece, predicted_value);
    // If both estimator and dfcm has not predicted then return
    if (!speculated && (!loaded))
    {
        InFlightInsL1.push(insMapping);
        return false;
    }
    else
    {
        // If estimator has not predicted then use the value predicted by dfcm
        if(!speculated)
        {
            insMapping.predicted_value = ld_predict;
            predicted_value = ld_predict;
        }
        // If estimator has predicted then we always use the value generated by estimator
        // because it has very less mispredictions
        else
            insMapping.predicted_value = predicted_value;

        insMapping.predicted = true;

        // Before predicting we verify if the PC is blacklisted
        // if it is blacklisted then we don't predict it
        std::set<uint64_t>::iterator wrong = BlacklistedPCTable.find(pc);
        if (wrong != BlacklistedPCTable.end())
        {
            InFlightInsL1.push(insMapping);
            return false;
        }

        InFlightInsL1.push(insMapping);
        return true;
    }
}

void speculativeUpdate(uint64_t seq_no, bool eligible, uint8_t prediction_result, uint64_t pc,
		       uint64_t next_pc, InstClass insn, uint8_t piece, uint64_t src1, uint64_t src2,
		       uint64_t src3, uint64_t dst)
{
    // we update dfcm with whatever update we have
    dfcm.speculativeUpdate(seq_no, eligible, prediction_result, pc, next_pc, insn, piece, src1, src2, src3, dst);

    // get inflight data
    InsMapping insMapping = InFlightInsL1.front();
    InFlightInsL1.pop();

    assert(insMapping.seq_no == seq_no);

    // Add the PC to the PCIT if not present
    addPCITEntry(pc, piece, src1, src2, src3, dst, insn, eligible);

    // We copy the register file entry from seq_no to seq_no + 1 so that
    // we might have got the output of seq_no and it should be reflected
    // in the register file entry of seq_no + 1
    copyRF(seq_no,seq_no + 1);

    // get the iterator of current PCs entry in PCIT
    std::unordered_map<uint64_t, PCITEntry>::iterator i = PCIT.find(SeqNotoPCMapper[seq_no].pc);

    // correct prediction
    if (prediction_result == 1)
    {
        // increase the score
        (i->second.score < MAX_SCORE) ?
            i->second.score++ : i->second.score;
    }
    // incorrect prediction
    else if (prediction_result == 0)
    {
        // decrease the score
        (i->second.score > -MAX_SCORE) ? i->second.score -= 5 : i->second.score;
        if (i->second.score <= THRESHOLD)
        {
            // add the PC to blacklist when score is less than the THRESHOLD value
            // getPrediction will return false on this PC till its score goes above THRESHOLD
            // the score will be updated at commit time and will increase if its predicted value matches with actual_value
            std::set<uint64_t>::iterator wrong = BlacklistedPCTable.find(pc);
            if (wrong == BlacklistedPCTable.end())
            {
                BlacklistedPCTable.insert(pc);
            }
        }

    }

    // update the register file with the predicted value if prediction is correct
    if (prediction_result == 1)
        updateRF(seq_no + 1, dst, insMapping.predicted_value, true, seq_no, piece);

    // for ineligible instruction and for wrong predictions update the register file and invalidate the dst entry
    if ((eligible == false || prediction_result != 1) && dst != 0xdeadbeef)
    {
        updateRF(seq_no + 1, dst, 0xdeadbeef, false, seq_no, piece);
    }

    // push the data to the next inflight queue for commit
    InFlightInsL2.push(insMapping);
}

void updatePredictor(uint64_t seq_no, uint64_t actual_addr,
		     uint64_t actual_value, uint64_t actual_latency)
{
    // update in dfcm also
    dfcm.updatePredictor(seq_no, actual_addr, actual_value, actual_latency);

    // pop the mapping of seq_no to PC from  the queue of inflight instructions
    PCITEntry pcitEntry;
    InsMapping insMapping = InFlightInsL2.front();
    InFlightInsL2.pop();
    curr_seq_no = seq_no;

    assert(insMapping.seq_no == seq_no);

    // get the entry in PCIT for the current PC
    // PC can be found from the insMapping structure
    std::unordered_map<uint64_t, PCITEntry>::iterator i = PCIT.find((insMapping.pc << 1) ^ insMapping.piece);
    //assert(i != PCIT.end());

    if(i == PCIT.end()) return;
    pcitEntry = i->second;

    // for ineligible instructions, just invalidate the dst register value in register file
    if (actual_value == 0xdeadbeef)
    {
        copyRF(seq_no, seq_no+1);
        // We do the copying again as the register file entry corresponding to seq_no might
        // have more updated values, this will potentially overwrite the updates that we made
        // to the RF in the speculative update but we update that again below
        if (pcitEntry.dst != 0xdeadbeef)
            updateRF(seq_no + 1, pcitEntry.dst, actual_value, false, seq_no, insMapping.piece);
        return;
    }

    // add or remove from the blacklist according to the score
    if (insMapping.predicted)
    {
        // correct prediction, increase the score and remove from blacklist if score is above threshold
        if (insMapping.predicted_value == actual_value)
        {
            (i->second.score < MAX_SCORE) ? i->second.score++ : i->second.score;

            if (i->second.score > THRESHOLD)
            {
                std::set<uint64_t>::iterator wrong = BlacklistedPCTable.find(insMapping.pc);
                if (wrong != BlacklistedPCTable.end())
                {
                    BlacklistedPCTable.erase(insMapping.pc);
                    space -= sizeof(uint64_t);
                }
            }
        }
        else
        {
            // incorrect prediction, decrease the score and add to blacklist if score is below threshold
            (i->second.score > -MAX_SCORE) ? i->second.score--:i->second.score;
            if (i->second.score <= THRESHOLD)
            {
                std::set<uint64_t>::iterator wrong = BlacklistedPCTable.find(insMapping.pc);
                if (wrong == BlacklistedPCTable.end() && space < MAX_SPACE)
                {
                    BlacklistedPCTable.insert(insMapping.pc);
                    space += sizeof(uint64_t);
                }
            }
        }
    }

    // update the register files accordingly
    copyRF(seq_no, seq_no+1);
    if (pcitEntry.dst != 0xdeadbeef)
        updateRF(seq_no + 1, pcitEntry.dst, actual_value, true, seq_no, insMapping.piece);

    // Compute the operation using estimator and update the pcitEntry
    OpData currOp = estimator(pcitEntry, RegisterFileBuffer[SeqNotoPCMapper[seq_no].index], actual_value);
    i->second.op = currOp.op;
    i->second.type = currOp.type;
    i->second.imm = currOp.imm;
    i->second.sure = currOp.sure;

    // Update the last data points
    i->second.lastdata1 = i->second.lastdata2;
    i->second.lastdata2 = i->second.lastdata3;
    Data curr_data;

    for (int i = 0; i < 3; i++)
    {
        if (pcitEntry.src[i] != 0xdeadbeef)
        {
            curr_data.src_val[i] = RegisterFileBuffer[SeqNotoPCMapper[seq_no].index][pcitEntry.src[i]].value;
        }
    }

    curr_data.dst_val = actual_value;
    i->second.lastdata3 = curr_data;
    i->second.lastdata3.valid = true;
}

// counts how many source register are used for a PC by comparing their values with deadbeef
uint8_t count_num_src(uint64_t src_val[])
{
    uint64_t num_src = 0;
    for (int i = 0; i < 3; i++)
    {
        if(src_val[i] != 0xdeadbeef)
            num_src++;
    }

    return num_src;
}

// Finds the amount of shift required to fit the given opcode and optype with data point
// try to match the opcode and optype by trying all the possible shift amounts
uint64_t shift_amount(OpCode op, OpType type, Data data)
{
    // range of shift amount is assumed to be 1-32
    for (uint64_t i = 0; i < 33; i++)
    {
        if (i<=31 && type == lshift_n && data.dst_val == calculate(op, type, data, i))
            return i;
        if (i && type == rshift_n && data.dst_val == calculate(op, type, data, i))
            return i;
    }
    return 0;
}

// checks if all the data point is consistent with operation
// use the op operation on the given data points to calculate the output and matches it with value in  destination register
bool consistency(OpCode op, OpType type, PCITEntry pcitEntry, uint64_t imm)
{
    // if the last data point is invalid, then PC must be occuring for the first time
    // so there is no data point for checking consistency, hence return true
    if (!pcitEntry.lastdata3.valid)
    {
        assert(!pcitEntry.lastdata1.valid && !pcitEntry.lastdata2.valid);
        return true;
    }

    // There is only one data point, hence check its consistency
    if (!pcitEntry.lastdata2.valid)
    {
        assert(!pcitEntry.lastdata1.valid);
        if(pcitEntry.lastdata3.dst_val == calculate(op, type, pcitEntry.lastdata3, imm))
            return true;
        else return false;
    }

    // check consistency of last two data points
    if (!pcitEntry.lastdata1.valid)
    {
        if((pcitEntry.lastdata2.dst_val == calculate(op, type, pcitEntry.lastdata2, imm)) && (pcitEntry.lastdata3.dst_val == calculate(op, type, pcitEntry.lastdata3, imm)))
            return true;
        else return false;
    }

    // check consistency of all the data points
    if((pcitEntry.lastdata1.dst_val == calculate(op, type, pcitEntry.lastdata1, imm)) && (pcitEntry.lastdata2.dst_val == calculate(op, type, pcitEntry.lastdata2, imm)) && (pcitEntry.lastdata3.dst_val == calculate(op, type, pcitEntry.lastdata3, imm)))
        return true;

    else return false;
}

// Find an operation that fits that PC.
// Verifies that operation with last 3 data points
// Returns OpData which contains opcode, optype, immediate value if present, and surity of that opcode
OpData estimator(PCITEntry pcitEntry, RegisterFileEntry currRegfile[], uint64_t actual_value)
{
    OpData currOp;
    currOp.op = None;
    currOp.type = none;
    currOp.imm = 0;
    currOp.sure = 0;

    // If it is already None, there is nothing we can do.
    // There must be a data point for this PC which doesn't fit to any of the instruction below
    if (pcitEntry.op == None)
    {
        return currOp;
    }

    // No need to estimate. Will use DFCM for prediction.
    if (pcitEntry.insn == loadInstClass)
    {
        currOp.op = ld;
        return currOp;
    }

    // No need to estimate. Will use DFCM for prediction.
    if (pcitEntry.insn == storeInstClass)
    {
        currOp.op = str;
        return currOp;
    }

    uint8_t num_src = count_num_src(pcitEntry.src);
    Data check_this_data;       // Current data point

    for (int i = 0; i < num_src; i++)
    {
        check_this_data.src_val[i] = currRegfile[pcitEntry.src[i]].value;
    }
    check_this_data.dst_val = actual_value;

    uint64_t imm = 0;

    // Varify with previous opcode
    // if matches, then increase the surity and return the same, otherwise try new opcodes and optypes
    if (pcitEntry.op != Init && consistency(pcitEntry.op, pcitEntry.type, pcitEntry, pcitEntry.imm) && actual_value == calculate(pcitEntry.op, pcitEntry.type, check_this_data, pcitEntry.imm))
    {
        currOp.op =  pcitEntry.op;
        currOp.type = pcitEntry.type;
        currOp.imm = pcitEntry.imm;
        currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
         return currOp;
     }

    // estimating the opertaion based on the number of source registers used
    switch(num_src)
    {
        case 0:
            {
                //mov imm
                // when the instruction is seen for the first time, the best guess is move the actual_value as immediate
                if(pcitEntry.op == Init)
                {
                    currOp.op = mov;
                    currOp.type = immed;
                    currOp.imm = actual_value;
                    currOp.sure = false;
                    return currOp;
                }
                //mov imm
                // for move instruction, all we have to do is to check consistency with older data.
                if(consistency(mov, immed, pcitEntry, actual_value))
                {
                    currOp.op = mov;
                    currOp.type = immed;
                    currOp.imm = actual_value;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //mvn imm
                if(consistency(mvn, immed, pcitEntry, ~actual_value))
                {
                    currOp.op = mvn;
                    currOp.type = immed;
                    currOp.imm = ~actual_value;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                return currOp;
            }
        case 1:
            {
                //mov reg
                if(consistency(mov, reg, pcitEntry, pcitEntry.imm) && actual_value == currRegfile[pcitEntry.src[0]].value)
                {
                    currOp.op = mov;
                    currOp.type = reg;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //mvn reg
                if(consistency(mvn, reg, pcitEntry, pcitEntry.imm) && actual_value == ~currRegfile[pcitEntry.src[0]].value)
                {
                    currOp.op = mvn;
                    currOp.type = reg;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //mov lshift_n
                // find n by brute force assuming mov instruction
                // if imm is 0, then this operation is not possible for this PC for any immediate shift value
                uint64_t imm = shift_amount(mov, lshift_n, check_this_data);
                if(imm && consistency(mov, lshift_n, pcitEntry, imm))
                {
                    currOp.op = mov;
                    currOp.type = lshift_n;
                    currOp.imm = imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //mov rshift_n
                imm = shift_amount(mov, rshift_n, check_this_data);
                if(imm && consistency(mov, rshift_n, pcitEntry, imm))
                {
                    currOp.op = mov;
                    currOp.type = rshift_n;
                    currOp.imm = imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //mvn lshift_n
                imm = shift_amount(mvn, lshift_n, check_this_data);
                if(imm && consistency(mvn, lshift_n, pcitEntry, imm))
                {
                    currOp.op = mvn;
                    currOp.type = lshift_n;
                    currOp.imm = imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //mvn rshift_n
                imm = shift_amount(mvn, rshift_n, check_this_data);
                if(imm && consistency(mvn, rshift_n, pcitEntry, imm))
                {
                    currOp.op = mvn;
                    currOp.type = rshift_n;
                    currOp.imm = imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //add immed
                // if we assume add instruction with an immediate value, then we check its consistency with all the last data points
                imm = actual_value - currRegfile[pcitEntry.src[0]].value;
                if(consistency(add, immed, pcitEntry, imm))
                {
                    currOp.op = add;
                    currOp.type = immed;
                    currOp.imm = imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //sub immed
                imm = currRegfile[pcitEntry.src[0]].value - actual_value;
                if(consistency(sub, immed, pcitEntry, imm))
                {
                    currOp.op = sub;
                    currOp.type = immed;
                    currOp.imm = imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //rsb immed
                imm = currRegfile[pcitEntry.src[0]].value + actual_value;
                if(consistency(rsb, immed, pcitEntry, imm))
                {
                    currOp.op = rsb;
                    currOp.type = immed;
                    currOp.imm = imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //_xor immed
                imm = currRegfile[pcitEntry.src[0]].value ^ actual_value;
                if(consistency(_xor, immed, pcitEntry, imm))
                {
                    currOp.op = _xor;
                    currOp.type = immed;
                    currOp.imm = imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                return currOp;
            }
        case 2:
            {
                // for register type instruction we have to verify the current data point and consistency with older data points
                //mov lshift_r
                if (consistency(mov, lshift_r, pcitEntry, pcitEntry.imm) && actual_value == (currRegfile[pcitEntry.src[0]].value << currRegfile[pcitEntry.src[1]].value)) {
                    currOp.op = mov;
                    currOp.type = lshift_r;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //mov rshift_r
                if (consistency(mov, rshift_r, pcitEntry, pcitEntry.imm) && actual_value == (currRegfile[pcitEntry.src[0]].value >> currRegfile[pcitEntry.src[1]].value)) {
                    currOp.op = mov;
                    currOp.type = rshift_r;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //mvn lshift_r
                if (consistency(mvn, lshift_r, pcitEntry, pcitEntry.imm) && actual_value == ~(currRegfile[pcitEntry.src[0]].value << currRegfile[pcitEntry.src[1]].value)) {
                    currOp.op = mvn;
                    currOp.type = lshift_r;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //mvn rshift_r
                if (consistency(mvn, rshift_r, pcitEntry, pcitEntry.imm) && actual_value == ~(currRegfile[pcitEntry.src[0]].value << currRegfile[pcitEntry.src[1]].value)) {
                    currOp.op = mvn;
                    currOp.type = rshift_r;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //add reg
                if (consistency(add, reg, pcitEntry, pcitEntry.imm) && actual_value == (currRegfile[pcitEntry.src[0]].value + currRegfile[pcitEntry.src[1]].value)) {
                    currOp.op = add;
                    currOp.type = reg;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //sub reg
                if (consistency(sub, reg, pcitEntry, pcitEntry.imm) && actual_value == (currRegfile[pcitEntry.src[0]].value - currRegfile[pcitEntry.src[1]].value)) {
                    currOp.op = sub;
                    currOp.type = reg;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //rsb reg
                if (consistency(rsb, reg, pcitEntry, pcitEntry.imm) && actual_value == (currRegfile[pcitEntry.src[1]].value - currRegfile[pcitEntry.src[0]].value)) {
                    currOp.op = rsb;
                    currOp.type = reg;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //_xor reg
                if (consistency(_xor, reg, pcitEntry, pcitEntry.imm) && actual_value == (currRegfile[pcitEntry.src[0]].value ^ currRegfile[pcitEntry.src[1]].value)) {
                    currOp.op = _xor;
                    currOp.type = reg;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //add lshift_n
                uint64_t imm = shift_amount(add, lshift_n, check_this_data);
                if (imm && consistency(add, lshift_n, pcitEntry, imm)) {
                    currOp.op = add;
                    currOp.type = lshift_n;
                    currOp.imm = imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //sub lshift_n
                imm = shift_amount(sub, lshift_n, check_this_data);
                if (imm && consistency(sub, lshift_n, pcitEntry, imm)) {
                    currOp.op = sub;
                    currOp.type = lshift_n;
                    currOp.imm = imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //rsb lshift_n 
                imm = shift_amount(rsb, lshift_n, check_this_data);
                if (imm && consistency(rsb, lshift_n, pcitEntry, imm)) {
                    currOp.op = rsb;
                    currOp.type = lshift_n;
                    currOp.imm = imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //_xor lshift_n 
                imm = shift_amount(_xor, lshift_n, check_this_data);
                if (imm && consistency(_xor, lshift_n, pcitEntry, imm)) {
                    currOp.op = _xor;
                    currOp.type = lshift_n;
                    currOp.imm = imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //add rshift_n
                imm = shift_amount(add, rshift_n, check_this_data);
                if (imm && consistency(add, rshift_n, pcitEntry, imm)) {
                    currOp.op = add;
                    currOp.type = rshift_n;
                    currOp.imm = imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //sub rshift_n
                imm = shift_amount(sub, rshift_n, check_this_data);
                if (imm && consistency(sub, rshift_n, pcitEntry, imm)) {
                    currOp.op = sub;
                    currOp.type = rshift_n;
                    currOp.imm = imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //rsb rshift_n 
                imm = shift_amount(rsb, rshift_n, check_this_data);
                if (imm && consistency(rsb, rshift_n, pcitEntry, imm)) {
                    currOp.op = rsb;
                    currOp.type = rshift_n;
                    currOp.imm = imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //_xor rshift_n 
                imm = shift_amount(_xor, rshift_n, check_this_data);
                if (imm && consistency(_xor, rshift_n, pcitEntry, imm)) {
                    currOp.op = _xor;
                    currOp.type = rshift_n;
                    currOp.imm = imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //mul
                if (consistency(mul, none, pcitEntry, pcitEntry.imm) && actual_value == (currRegfile[pcitEntry.src[0]].value * currRegfile[pcitEntry.src[1]].value))
                {
                    currOp.op = mul;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                return currOp;
            }
        case 3:
            {
                //add lshift_r
                if (consistency(add, lshift_r, pcitEntry, pcitEntry.imm) && actual_value == calculate(add, lshift_r, check_this_data, pcitEntry.imm))
                {
                    currOp.op = add;
                    currOp.type = lshift_r;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //sub lshift_r
                if (consistency(sub, lshift_r, pcitEntry, pcitEntry.imm) && actual_value == calculate(sub, lshift_r, check_this_data, pcitEntry.imm))
                {
                    currOp.op = sub;
                    currOp.type = lshift_r;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //rsb lshift_r
                if (consistency(rsb, lshift_r, pcitEntry, pcitEntry.imm) && actual_value == calculate(rsb, lshift_r, check_this_data, pcitEntry.imm))
                {
                    currOp.op = rsb;
                    currOp.type = lshift_r;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //_xor lshift_r
                if (consistency(_xor, lshift_r, pcitEntry, pcitEntry.imm) && actual_value == calculate(_xor, lshift_r, check_this_data, pcitEntry.imm))
                {
                    currOp.op = _xor;
                    currOp.type = lshift_r;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //add rshift_r
                if (consistency(add, rshift_r, pcitEntry, pcitEntry.imm) && actual_value == calculate(add, rshift_r, check_this_data, pcitEntry.imm))
                {
                    currOp.op = add;
                    currOp.type = rshift_r;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //sub rshift_r
                if (consistency(sub, rshift_r, pcitEntry, pcitEntry.imm) && actual_value == calculate(sub, rshift_r, check_this_data, pcitEntry.imm))
                {
                    currOp.op = sub;
                    currOp.type = rshift_r;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //rsb rshift_r
                if (consistency(rsb, rshift_r, pcitEntry, pcitEntry.imm) && actual_value == calculate(rsb, rshift_r, check_this_data, pcitEntry.imm))
                {
                    currOp.op = rsb;
                    currOp.type = rshift_r;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //_xor rshift_r
                if (consistency(_xor, rshift_r, pcitEntry, pcitEntry.imm) && actual_value == calculate(_xor, rshift_r, check_this_data, pcitEntry.imm))
                {
                    currOp.op = _xor;
                    currOp.type = rshift_r;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                //mla
                if (consistency(mla, none, pcitEntry, pcitEntry.imm) && actual_value == calculate(mla, none, check_this_data, pcitEntry.imm))
                {
                    currOp.op = mla;
                    currOp.imm = pcitEntry.imm;
                    bool correct = ((currOp.op == pcitEntry.op) && (currOp.type == pcitEntry.type));
                    if(correct) currOp.sure = (pcitEntry.sure < 2)?pcitEntry.sure+1:pcitEntry.sure;
                    else currOp.sure = 0;
                    return currOp;
                }
                return currOp;
            }
        default:
                return currOp;

    }
    return currOp;
}

// Takes the opcode, optype and data point as input and computes the value
uint64_t calculate(OpCode op, OpType type, Data data, uint64_t imm)
{
    if (op == mov) {
        if(type == immed)
        {
            return imm;
        }
        if(type == reg)
        {
            return data.src_val[0];
        }
        if(type == lshift_n)
        {
            return data.src_val[0] << imm;
        }
        if(type == rshift_n)
        {
            return data.src_val[0] >> imm;
        }
        if(type == lshift_r)
        {
            return data.src_val[0] << data.src_val[1];
        }
        if(type == rshift_r)
        {
            return data.src_val[0] >> data.src_val[1];
        }
    }
    if (op == mvn) {
        if(type == immed)
        {
            return ~imm;
        }
        if(type == reg)
        {
            return ~(data.src_val[0]);
        }
        if(type == lshift_n)
        {
            return ~(data.src_val[0] << imm);
        }
        if(type == rshift_n)
        {
            return ~(data.src_val[0] >> imm);
        }
        if(type == lshift_r)
        {
            return ~(data.src_val[0] << data.src_val[1]);
        }
        if(type == rshift_r)
        {
            return ~(data.src_val[0] >> data.src_val[1]);
        }
    }
    if (op == add) {
        if(type == immed)
        {
            return data.src_val[0] + imm;
        }
        if(type == reg)
        {
            return data.src_val[0] + data.src_val[1];
        }
        if(type = lshift_n)
        {
            return data.src_val[0] + (data.src_val[1] << imm);
        }
        if(type = rshift_n)
        {
            return data.src_val[0] + (data.src_val[1] >> imm);
        }
        if(type = lshift_r)
        {
            return data.src_val[0] + (data.src_val[1] << data.src_val[2]);
        }
    }
    if (op == sub) {
        if(type == immed)
        {
            return data.src_val[0] - imm;
        }
        if(type == reg)
        {
            return data.src_val[0] - data.src_val[1];
        }
        if(type = lshift_n)
        {
            return data.src_val[0] - (data.src_val[1] << imm);
        }
        if(type = rshift_n)
        {
            return data.src_val[0] - (data.src_val[1] >> imm);
        }
        if(type = lshift_r)
        {
            return data.src_val[0] - (data.src_val[1] << data.src_val[2]);
        }
    }
    if (op == rsb) {
        if(type == immed)
        {
            return imm - data.src_val[0];
        }
        if(type == reg)
        {
            return data.src_val[1] - data.src_val[0];
        }
        if(type = lshift_n)
        {
            return (data.src_val[1] << imm) - data.src_val[0];
        }
        if(type = rshift_n)
        {
            return (data.src_val[1] << imm) - data.src_val[0];
        }
        if(type = lshift_r)
        {
            return  (data.src_val[1] << data.src_val[2]) - data.src_val[0];
        }
    }
    if (op == _xor) {
        if(type == immed)
        {
            return data.src_val[0] ^ imm;
        }
        if(type == reg)
        {
            return data.src_val[0] ^ data.src_val[1];
        }
        if(type = lshift_n)
        {
            return data.src_val[0] ^ (data.src_val[1] << imm);
        }
        if(type = rshift_n)
        {
            return data.src_val[0] ^ (data.src_val[1] >> imm);
        }
        if(type = lshift_r)
        {
            return data.src_val[0] ^ (data.src_val[1] << data.src_val[2]);
        }
    }
    if (op == mul) {
        if(type == none)
        {
            return data.src_val[0] * data.src_val[1];
        }
    }
    if (op == mla) {
        if(type == none)
        {
            return (data.src_val[0] * data.src_val[1]) + data.src_val[2];
        }
    }
    return 0;
}

void beginPredictor(int argc_other, char **argv_other)
{
    // Initialising registerfile with all invalid entries
    for (int j = 0; j < MAX_SEQ_BUFFER; j++)
    {
        RFBIndextoSeqNoMapper[j] = ULONG_MAX;
        for (int i = 0; i < MAX_REGS; i++)
        {
            RegisterFileBuffer[j][i].valid = false;
            RegisterFileBuffer[j][i].value = 0;
            RegisterFileBuffer[j][i].seq_no = 0;
        }
    }

    if (argc_other > 0)
        printf("CONTESTANT ARGUMENTS:\n");

    for (int i = 0; i < argc_other; i++)
        printf("\targv_other[%d] = %s\n", i, argv_other[i]);
}

void endPredictor() {
    printf("CONTESTANT OUTPUT--------------------------\n");
    printf("--------------------------\n");
    printf("size of structures:-\n");
    printf("Size of firstLevelTable: %ld\n", (sizeof(FirstLevelEntry)+8)*dfcm.firstLevelTable.size());
    printf("Size of secondLevelTable0: %ld\n", (sizeof(SecondLevelEntry)+8)*dfcm.secondLevelTable[0].size());
    printf("Size of secondLevelTable1: %ld\n", (sizeof(SecondLevelEntry)+8)*dfcm.secondLevelTable[1].size());
    printf("Size of DFCM inflight Preds: %ld\n", (sizeof(DFCM::InflightInfo))*dfcm.inflightPreds.size());
    printf("Size of PCIT: %ld\n", (sizeof(PCITEntry)+8)*PCIT.size());
    printf("Size of SeqNotoPCMapper: %ld\n", (sizeof(SeqNoEntry)+8)*SeqNotoPCMapper.size());
    printf("Size of BlacklistedPCTable: %ld\n", (8)*BlacklistedPCTable.size());
    printf("Size of inflight Preds: %ld\n", (sizeof(InsMapping))*(InFlightInsL1.size()+ InFlightInsL2.size()));
    printf("space used: %ld\n", space);
}
