#include <unordered_map>
#include <set>
#include <queue>
#include <vector>
#include "cvp.h"

#define TABLES 2                // Number of different context lenghtsr, currently one is 32 and other is 64.
#define MAX_REGS 65             // Max number of registers in register file.
#define MAX_SEQ_BUFFER 500      // Max limit of inflight instructions.

// Structure for storing the context.
struct __attribute__((packed)) FirstLevelEntry 
{
    uint64_t commitTimeIndex[TABLES];   // Hash of last N values
    uint64_t predictTimeIndex[TABLES];  // speculative Hash of last N values
    uint64_t inflight;                  // Number of inflight instructions corresponding to this entry
    uint64_t baseValue;                 // corresponds to the last seen value for this entry 
    uint64_t predictValue;              // speculative base value
    bool predictValid;                  // represents the validity of predictValue
                                        // if false then we don't predict using speculative value
    int preference[TABLES];             // preference counter (3bit*2) to select between different context lengths
                                        // one for each table
    int maxConf;                        // Dynamic confidence threshold for each PC, we only use 5 bits of this counter

    // Entry size: 460 bits 

    // Constructor to initialize the struct variables by 0.
    FirstLevelEntry() : inflight(0), baseValue(0), predictValue(0), predictValid(true), maxConf(10) {
        for (int i = 0; i < TABLES; i++) {
            predictTimeIndex[i] = 0;
            commitTimeIndex[i] = 0;
            preference[i] = 0;
        }
    }
};
  
struct __attribute__((packed)) SecondLevelEntry
{
    int64_t pred;   // predicted value of stride
    uint8_t conf;   // confidence of entry, we only need 4 bits of this

    // Entry size: 68 bits

    // Constructor to initialize the struct variables by 0.
    SecondLevelEntry() : pred(0), conf(0) {};
};
  
// This structure stores the information for the DFCM tables and context.
struct DFCM
{
    // Stores the context.  
    // This corresponds to the SHT structure
    // It correpsonds to row 1 of hardware overhead table
    std::unordered_map<uint64_t, FirstLevelEntry> firstLevelTable;

    // We have to maintain N second level tables corresponding to each context length
    // This corresponds to row 2 of hardware overhead table
    std::unordered_map<uint64_t, SecondLevelEntry> secondLevelTable[TABLES]; // refered to as SPT in paper
  
    // structure to store information about in flight instructions
    // Part of unlimited metadata buffer
    struct InflightInfo
    {
        uint64_t seqNum;
    	uint64_t firstLevelIndex;
    	uint64_t secondLevelIndex[TABLES];
	
    	InflightInfo() : seqNum(0), firstLevelIndex(0) {} 
    	InflightInfo(uint64_t sn, uint64_t fidx, uint64_t sidx[]) : seqNum(sn), firstLevelIndex(fidx) {
            for (int i = 0; i < TABLES; i++) {
                secondLevelIndex[i] = sidx[i];
            }
        }
    };
  
    // The queue which stores the details of inflight instructions
    std::deque<InflightInfo> inflightPreds;
  
    int64_t lastPrediction;   // The last prediction
    uint64_t lastStride;      // The stride used for last prediction

    // Methods of DFCM to update the predictor, this allows dfcm to act as a independent 
    // entity which gets its information via estimator
    bool getPrediction(uint64_t seq_no, uint64_t pc, uint8_t piece, uint64_t& predicted_value);
    void speculativeUpdate(uint64_t seq_no, bool eligible, uint8_t prediction_result, uint64_t pc,			
		           uint64_t next_pc, InstClass insn, uint8_t piece, uint64_t src1, uint64_t src2,
		           uint64_t src3, uint64_t dst);
    void updatePredictor(uint64_t seq_no, uint64_t actual_addr,
		         uint64_t actual_value, uint64_t actual_latency);

public:
	
    DFCM()
    {
    	lastPrediction = 0xdeadbeef;
        lastStride = 0;
    }
};


enum __attribute__((packed)) OpCode {
    // all the PC will be intialized with Init opcode
    // no prediction will be done by estimator for none type instruction
    // rsb is reverse subtract
    // mvn is move negative
    // mla is multiply and  add
    None, Init, add, sub, rsb, mul, _xor, mov, mvn, mla, ld, str
};

enum __attribute__((packed)) OpType {
    // n represent shift by immediate value
    // r represent shift by register value
    // reg represent that operation is purely register type without any shift
    // immed represent that operation is immediate type
    none, immed, reg, lshift_n, rshift_n, lshift_r, rshift_r
};

// Structure to store a data point corresponding to a instruction
// Entry size: 257 bits 
struct __attribute__((packed)) Data {
    bool valid;             // validity of the data point
    uint64_t src_val[3];    // value of the source registers
    uint64_t dst_val;       // value of the destination register
};


// Structure to store per PC information
// Entry size: 1112 bits
// This corresponds to the row 3 in hardware overhead table
struct __attribute__((packed)) PCITEntry {
    uint64_t src[3];    // source register numbers
    uint64_t dst;       // destination register number
    uint64_t imm;       // immediate value for that PC
    OpCode op;          // This takes 4 bits
    OpType type;        // This also takes 4 bits
    uint64_t sure;      // surity of the opcode, only 2 bits are used
    bool eligible;      // eligibility of the PC
    InstClass insn;     // This is defined by cvp.h and can be stored in 4 bits
    Data lastdata1;     // last 3rd data point
    Data lastdata2;     // last 2nd data point
    Data lastdata3;     // last data point
    int score;          // saturating counter for blacklisting (6 bits)
};

// Structure used to store information about each register 
// Entry size: 193 bits
// This corresponds to the row 4 in hardware overhead table
struct __attribute__((packed)) RegisterFileEntry {
    bool valid;         // indicates whether the value is valid or stale
    uint64_t value;     // the register value
    uint64_t seq_no;    // seq_no of the last ins that updated the register
                        // it can only increase as we don't allow older instructions
                        // to write to a register which has been updated by the a new
                        // instruction
    uint8_t piece;     // the piece of the instruction which updated the entry
};

// Structure used to transfer data between getPrediction/speculativeUpdate/updatePredictor
// Part of unlimited meta buffer 
struct __attribute__((packed)) InsMapping {
    uint64_t pc;
    uint8_t piece;
    uint64_t seq_no;
    bool predicted;
    uint64_t predicted_value;
};

// Structure to store the behaviour of instruction
// Doesn't corresponds to a hardware component, used to easily
// carry data across functions
struct __attribute__((packed)) OpData {
    OpCode op;
    OpType type;
    uint64_t imm;
    uint64_t sure;
};

// Stores the instance of register file at different sequence numbers.
// We only need to store the register file state for inflight instrucions
// This is part of the unlimited meta buffer
// This corresponds to the row 5 in hardware overhead table
struct RegisterFileEntry RegisterFileBuffer[MAX_SEQ_BUFFER][MAX_REGS];

// Structure to map seq_no to index in the register file buffer
struct __attribute__((packed)) SeqNoEntry {
    uint64_t pc;
    int index; // index in the register file
};

// Structures used for maintaining the register file buffer
// of inflight instructions, also part of the unlimited buffer
// For mapping between PC and seq_no.
// This corresponds to the row 6 in hardware overhead table
std::unordered_map<uint64_t, struct SeqNoEntry> SeqNotoPCMapper;
// This corresponds to the row 7 in hardware overhead table
uint64_t RFBIndextoSeqNoMapper[MAX_SEQ_BUFFER];

// Stores information of a PC.
// Entry size: 1201 bits
// This corresponds to the row 3 of hardware overhead table 
std::unordered_map<uint64_t, PCITEntry> PCIT;

// Blacklist table.
// This corresponds to the row 8 in hardware overhead table
std::set<uint64_t> BlacklistedPCTable;

// we have 2 queues to transfer data between 3 functions (getPrediction -> speculativeUpdate -> updatePredictor)
// Part of the unlimited metadata buffer
std::queue <InsMapping> InFlightInsL1;
std::queue <InsMapping> InFlightInsL2;

// Helper function to count the number of valid src values
uint8_t count_num_src(uint64_t src_val[]);

// Estimating function.
OpData estimator(PCITEntry pcitEntry, RegisterFileEntry currRegfile[], uint64_t actual_value);

// Calculates the value for a given opcode and optype.
uint64_t calculate(OpCode op, OpType type, Data data, uint64_t imm);

/* The src and dst are register indexes */
// Adds a new entry in PCITEntryTable if not present.
void addPCITEntry(uint64_t pc, uint8_t piece, uint64_t src1, uint64_t src2, uint64_t src3, uint64_t dst, InstClass insn, bool eligible);

// For updating and maintaining register file.
int addRF(uint64_t seq_no);
void copyRF(uint64_t src, uint64_t dst);
void updateRF(uint64_t reg, uint64_t value, uint64_t seq_no);

// Use dependency graph to predict the value.
bool speculate(uint64_t seq_no, uint8_t piece, uint64_t& predicted_value);

// for checking the opcode consistency with all the data points of that PC.
bool consistency(OpCode op, OpType type, PCITEntry pcitEntry, uint64_t imm);

// Finds the immediate shift amount for and operation.
uint64_t shift_amount(OpCode op, OpType type, Data data);
